""" This is a dummy file used only to avoid errors in ReadTheDocs. The real BF.py is created during the setup once swig is run. """


def CP():
    pass


def LeP():
    pass


def LaP():
    pass


def HoPpro():
    pass


def HoPphy():
    pass


def FS():
    pass


def ELMReLU():
    pass


def ELMSigmoid():
    pass


def ELMTanh():
    pass


def ELMSin():
    pass


def ELMSwish():
    pass


def nCP():
    pass


def nLeP():
    pass


def nFS():
    pass


def nELMReLU():
    pass


def nELMSigmoid():
    pass


def nELMTanh():
    pass


def nELMSin():
    pass


def nELMSwish():
    pass
